#!/usr/bin/env python3
#
# mmgen = Multi-Mode GENerator, command-line Bitcoin cold storage solution
# Copyright (C)2013-2024 The MMGen Project <mmgen@tuta.io>

"""
test.objtest_py_d.ot_common: shared data for MMGen data objects tests
"""

from ..include.common import getrand

r16,r32 = getrand(16),getrand(32)
